#!/usr/bin/env tclsh
package require Tk
namespace path ::tcl::mathop

set hilfe {cadspline.tcl dient zur Formung von Eckenverrundungen zwecks
 Import in Librecad oder in Openscad -> hier als polygon([[]....[]]); <-.
 Die Routine zur Punkteberechnung habe ich mir selbst einfallen lassen, 
 vielleicht ist sie in der Literatur im Umfeld der Stichworte 'Quadratische Splines' 
 oder 'Chaikin-Algorithmus' wiederzufinden. Der Mechanismus der Routine 
 lässt sich ohne Weiteres auf höhere Dimensionen als nur 2 anwenden, z.B. 3D + 3 Farben.
 cadspline.tcl muss ausführbar gemacht werden, wenn der Aufruf mittels 
 "tclsh cadspline.tcl" vermieden werden soll. Das Programm läuft bei mir unter TclTk 8.6.
 Es bietet ein Textfenster und eine Buttonleiste. Hier können die Polygonrohdaten 
 in der Form "x1 y1 x2 y2 ........ xn yn" (mit Dezimalpunkten) eingegeben werden, 
 wobei Trennung mit RET der Übersicht dient - auf Komfortfunktionen habe ich verzichtet. 
 Die eingegebenen Rohwerte können mit 'Save' gespeichert, mit 'Neu' gelöscht 
 und mit 'Load' wiederhergestellt werden.
 In der Button-Leiste sind zwei editierbare Spinboxes 'Linear' sowie
 'Glaette' untergebracht. Mittels 'Linear' kann gewählt werden, wie groß
 der mittige Anteil eines jeden Polygonabschnittes ist, der nicht in 
 die Krümmung einbezogen wird - Vorwahl ist das Minimum von 0.0, was für 
 'Krümmung über die gesamte Kante' steht, Maximum ist 1.0, was für das 
 Fehlen jeder Krümmung steht. Gerade Abschnitte in der Kontur können natürlich 
 auch durch explizite Angabe in der Punkteabfolge erzwungen werden.
 Mit der Spinbox 'Glätte' kann die Glätte des entstehenden Linienzugs 
 beeinflusst werden - 0 ergibt die Originalpunkte.
 Sind diese Einstellungen erledigt, wird mittels des Buttons 'LibreCad' eine
 Punktfolge konstruiert, die aus LibreCAD über das Plugin 'Read ascii points' 
 mit Einstellung 'space separator' verwendbar ist.
 Entsprechend wird mittels 'Openscad' eine OpenSCAD-Datei generiert, die mit 
 '<include...>' in OpenSCAD als fertiges Polygon importiert werden kann.
 Unter Ausschluss jeglicher Gewährleistung entlasse ich das Programmfragment in die Public Domain
 
 Köln am 19.07.19 
 ha-schreiber@web.de
 H. Schreiber
}

set soviele {0 1 2 3 4 5 6 7 8}
set zpunkte 0
set luecke 0
set lueckenwerte {0.0 0.1 0.2 0.25 0.3333 0.4 0.5 0.6 0.6667 0.7 0.75 0.8 0.9 1.0}



frame .tx
text .tx.text -wrap none -yscrollcommand ".tx.vscroll set" -width 100 -height 20
scrollbar .tx.vscroll -command ".tx.text yview"
pack .tx.vscroll -side right -fill y
pack .tx.text -side left
pack .tx -side top

frame .bt -bg darkgray
button .bt.load -text Load -command laden -bg ivory
button .bt.save -text Save -command speichern -bg ivory
button .bt.libre -text LibreCad -command librecad -bg green
button .bt.scad -text Openscad -command openscad -bg yellow
label .bt.lbz -text " Linear:" -bg darkgray
spinbox .bt.sbz -relief sunken -textvariable luecke -values $lueckenwerte -width 5 -bg ivory 
label .bt.lb -text " Glätte:" -bg darkgray
spinbox .bt.sb -relief sunken -textvariable zpunkte -values $soviele -width 4 -bg ivory -state readonly
button .bt.hlp -relief raised -text Hilfe -bg lightblue -command {
	global hilfe
	toplevel .hilfe 
	frame .hilfe.fr
	text .hilfe.fr.txt -width 100 -height 35 -bg lightblue -wrap word
	button .hilfe.fr.bt -text Fertig -command "destroy .hilfe"
	pack .hilfe.fr.txt -side top -fill x
	pack .hilfe.fr.bt -side top -fill x
	pack .hilfe.fr -side top -fill x
	.hilfe.fr.txt insert end $hilfe
	.hilfe.fr.txt configure -state disabled
	}
	
	

	
button .bt.neu -text Neu -command {.tx.text delete 1.0 end} -bg orange 
button .bt.ex -text EXIT -command exit -bg red

pack .bt.load -side left -padx 1 -pady 2 
pack .bt.save -side left -padx 1 -pady 2
pack .bt.libre -side left -padx 1 -pady 2
pack .bt.scad -side left -padx 1 -pady 2
pack .bt.lbz -side left -padx 1 -pady 0
pack .bt.sbz -side left -padx 5 -pady 0
pack .bt.lb -side left -padx 1 -pady 0
pack .bt.sb -side left -padx 5 -pady 0

pack .bt.ex -side right -padx 1 -pady 2
pack .bt.neu -side right -padx 1 -pady 2
pack .bt.hlp -side right -padx 1 -pady 2
pack .bt -side top -fill x

# einige Hilfsfunktionen in 2D
# zur späteren Verwendung

proc mitte {A B} {
	# der Mittelpunkt res zweier gegebener Punkte A und B in 2D
	set res {}
	lassign $A ax ay
	lassign $B bx by
	lappend res [/ [+ $ax $bx] 2.0]
	lappend res [/ [+ $ay $by] 2.0]
	return $res
}

proc plus {A B} {
	return [list [+ [lindex $A 0] [lindex $B 0]] [+ [lindex $A 1] [lindex $B 1]]]
}

proc minus {A B} {
	return [list [- [lindex $A 0] [lindex $B 0]] [- [lindex $A 1] [lindex $B 1]]]
}

proc mal {A wert} {
	set u [* [lindex $A 0] $wert]
	set v [* [lindex $A 1] $wert]
	return [list $u $v]
}



proc gleich {A B} {
	if {[lindex $A 0] == [lindex $B 0] && [lindex $A 1] == [lindex $B 1]} {
		return 1
	}
	return 0
}


proc scheitel {A C B} {
	# scheitel {A B C} ermittelt den Scheitelpunkt einer der möglichen 
	# in A zu AC und in B zu BC genau tangentialen Kurven im Inneren
	# von ABC
	return [mitte [mitte $A $C] [mitte $B $C]]
}



	
# Ende der Hilfsfunktionen

proc laden {} {
	set name [tk_getOpenFile -title Quelldatei]
	if {![file readable $name]} return
	if {[catch {set fd [open $name r]} fehler]} {
		puts $fehler
		return
	}
	
	
	set inhalt [read -nonewline $fd]
	close $fd
	.tx.text delete 1.0 end
	.tx.text insert 1.0 $inhalt
}

proc speichern {} {
	set name [tk_getSaveFile]
	if {[catch {set fd [open $name w]} fehler]} {
		puts $fehler
		return
	}
	puts -nonewline $fd [.tx.text get 1.0 end]
	close $fd
}

proc prepare {werte} {
	global luecke
	
	set lw [llength $werte]
	
	
	# bei den weiteren Berechnungen werden NichtInteger erzwungen
	for {set i 0} {$i < $lw} {incr i} {
		lset werte $i [* [lindex $werte $i ] 1.0]
	}
	
	# Die XY-Koordinaten werden als Liste von 2D-Punkten verpackt
	set paare {}
	while {[llength $werte] > 1} {
		lappend paare [lrange $werte 0 1]
		set werte [lreplace $werte 0 1]
	}
	
		
	# die Punktekette muss offen sein, damit bei der Berechnung der 
	# Zwischenwerte der Ringschluss funktioniert
	
	if {[gleich [lindex $paare 0] [lindex $paare end]]} {
		set paare [lreplace $paare end end]
	}
	
	
	
	
	# die Endpunkte der Geradenabschnitte im gerundeten 
	# Polygonzug werden eingefügt
	# zluecke wird aus luecke berechnet
	
	
	set zluecke [/ [- 1.0 $luecke] 2.0]
	
	set lp [llength $paare]	
	set i [- $lp 1]
	while {$i > 0}  {
		set k $i
		incr i -1
		set p2 [lindex $paare $k]
		set p1 [lindex $paare $i]
		set intervall [minus $p2 $p1]
		set abstand [mal $intervall $zluecke]
		set paare [linsert $paare $k [plus $p1 $abstand]]
		incr k
		set paare [linsert $paare $k [minus $p2 $abstand]]
		#incr i -1
	}
	
	# schließlich muessen noch zwei freie Linearitätsenden eingefügt werden
	
	set p2 [lindex $paare 0]
	set p1 [lindex $paare end]	
	set intervall [minus $p2 $p1]
	set abstand [mal $intervall $zluecke]
	set paare [linsert $paare 0 [minus $p2 $abstand]]
	lappend paare [plus $p1 $abstand]
	
	return $paare
}

proc biegung {werte} {
	global luecke zpunkte
	
	if {$luecke > 1.0 || $luecke < 0.0} {
		puts "Uzulässiger Wert für luecke"
		return $werte
		}
	
	set paare [prepare $werte]
	set bogen {}
	
	set sprung 6
	set beginn 0
	for {set t 0} {$t < $zpunkte} {incr t} {
		
		set anzahl [/ [llength $paare] 3]
		
		for {set i 0} {$i < $anzahl} {incr i} {
			set A [lindex $paare 0]
			set C [lindex $paare 1]
			set B [lindex $paare 2]
			lappend paare $A [mitte $A $C] [scheitel $A $C $B] [scheitel $A $C $B] [mitte $C $B] $B
			set paare [lreplace $paare 0 2]		
		}
	}
	
	while {[llength $paare] > 2} {
		set paare [lassign $paare A C B]
		set bogen [concat $bogen $A $C]		
	}
	set bogen [concat $bogen $B]
	
	
return $bogen	
}
		
	
	

proc quadspline {} {
	global zpunkte luecke
	
	
	set seite [split [.tx.text get 1.0 end]]
	set woerter {}
	foreach wort $seite {
		if {[llength $wort] < 1} {continue}
		lappend woerter $wort
	}
	
	set llw [llength $woerter]
	if {$llw < 6 || [% $llw 2] == 1} {
		puts "Zu wenige oder unpaarige Koordinaten"
		return {}
	}
	
	# Einige Regelverletzungen führen zu Nichtbeachtung
	if {$luecke >= 1.0 || $luecke < 0.0 || $zpunkte == 0} {
		return $woerter
	}
	
	return [biegung $woerter]
	
}
	

proc librecad {} {
	set woerter [quadspline]
	set zaehler 1
	set results {}
	while {[llength $woerter] > 1} {
		lappend results "$zaehler [lindex $woerter 0] [lindex $woerter 1]"
		set woerter [lreplace $woerter 0 1]	
		incr zaehler	
	}
	
	if {[llength $results] < 1} {
		puts "Keine Ergebnisse von quadspline"
		return
	}
	
	set name [tk_getSaveFile]
	if {[catch {set fd [open $name w]} fehler]} {
		puts $fehler
		return
	}
	foreach zeile $results {
		puts $fd $zeile
		
	}
	close $fd
	
}


proc openscad {} {
	set woerter [quadspline]
	set ergebnis "(\["
	while {[llength $woerter] > 1} {
		append ergebnis "\n"
		append ergebnis "\[[lindex $woerter 0],[lindex $woerter 1]\],"
		set woerter [lreplace $woerter 0 1]
	}	
	set ergebnis "polygon[string trim $ergebnis ,]\])"
	append ergebnis ";"
	
	set name [tk_getSaveFile -defaultextension .scad]
	if {[catch {set fd [open $name w]} fehler]} {
		puts $fehler
		return
	}
	puts $fd $ergebnis
	close $fd
}

		


	
	


